local B, L, T = unpack(select(2, ...))

--[[//////////////////////////////////////////////////

    HELPER FUNCTIONS

    Anything that needs repetitive actions or calculations
    Tips: 
        -- Multiplication are faster than division
        -- Local is faster than global. Re-assign global as a new local if used a lot

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    -- Table pool for recycling tables
    local tables = setmetatable({}, {__mode = "kv"}) -- Weak table

    -- Get a new table
    function B.newtable()
        local t = next(tables) or {}

        tables[t] = nil
        return t
    end

    -- Delete table and add to table pool
    function B.deltable(t)

        if type(t) == "table" then
            for k, v in pairs(t) do
                if type(v) == "table" then B.deltable(v) end
                t[k] = nil
            end
            
            setmetatable(t, nil)
            t[true] = true
            t[true] = nil
            tables[t] = true
        end
        return nil
    end

    -- tcopy: recursively copy contents of one table to another.  from wowwiki
    function B.copytable(to, from)
        -- "to" must be a table (possibly empty)
        for k, v in pairs(from) do

            if type(v) == "table" then
                to[k] = {}
                B.copytable(to[k], v)
            else
                to[k] = v
            end
        end
    end

    function B.template(server, player, faction)
        local template

        B.print(T.MSGC.EVENT.."Generating template: "..player)
        T.ACTIVE_TABLE = {['server'] = server, ['player'] = player, ['faction'] = faction}

        template = {['bank'] = {[1] = {['slots'] = NUM_BANKGENERIC_SLOTS}}, ['reagent'] = {},
                    ['void'] = {}, ['guild'] = {}}
        
        if not BWsavedtable then
            BWsavedtable = {}
            BWsavedtable['iconsizeslider'] = 32
            BWsavedtable['iconsize'] = 32
            BWsavedtable['iconrowslider'] = 12
            BWsavedtable['iconrow'] = 12
            BWsavedtable['minimize'] = false
            BWsavedtable['helptooltip'] = true
        end

        if not BWsavedtable[server] then BWsavedtable[server] = {} end

        BWsavedtable[server][player.."."..string.sub(faction, 1, 1)] = template
    end

    function B.load(name, player, faction)
        local text

        T.ACTIVE_TABLE = {['server'] = name, ['player'] = player, ['faction'] = faction}
        -- Validate settings added at a later time
        if BWsavedtable['helptooltip'] == nil then BWsavedtable['helptooltip'] = true end

        T.BW.items = B.deltable(T.BW.items) or B.newtable()

        B.BankUpdate()
        B.ReagentUpdate()
        B.VoidUpdate()
        B.GuildUpdate()

        text = T.F.mainframe.searchbar:GetText()
        if text then T.F.mainframe.searchbar:SetText(""); T.F.mainframe.searchbar:SetText(text) end
    end

    function B.GetPlayerBank(bank, current)

        if not bank then return end
        if current then
            local player = UnitName("player")
            local server = GetRealmName()
            local faction = UnitFactionGroup("player")
            return BWsavedtable[server][player.."."..string.sub(faction, 1, 1)][bank]
        end

        return BWsavedtable[T.ACTIVE_TABLE.server][T.ACTIVE_TABLE.player.."."..string.sub(T.ACTIVE_TABLE.faction, 1, 1)][bank]
    end


    local stack = -1
    local messagestr = ""

    function B.print(message, level)
        local filler

        if BWdebug or (level == 1) then

            if stack == -1 then messagestr = T.MSGC.DEFAULT.."Bank's Watch|r: " end

            if not message then
                DEFAULT_CHAT_FRAME:AddMessage(messagestr..tostring(message))
                return false 
            end

            stack = stack + 1

            if type(message) == "table" then
                
                filler = string.rep(". . ", stack)
                DEFAULT_CHAT_FRAME:AddMessage(messagestr.."{table} -->")
                for k, v in pairs(message) do
                    messagestr = filler.."["..k.."] = "
                    B.print(v)
                end
            
            elseif type(message) == "userdata" then

            elseif type(message) == "function" then
                DEFAULT_CHAT_FRAME:AddMessage(messagestr.." A function(self)")
            elseif type(message) == "boolean" then
                DEFAULT_CHAT_FRAME:AddMessage(messagestr..tostring(message))
            else
                DEFAULT_CHAT_FRAME:AddMessage(messagestr..tostring(message))
            end
            
            stack = stack - 1
       end
    end

    function B.find(name, text)

        if name and string.find(string.lower(string.gsub(name, "[%-]", " ")), string.lower(string.gsub(text, "[%-]", " "))) then
            return true
        end
        return false
    end

    function B.GetCursorScaledPosition()
        local scale, x, y = UIParent:GetScale(), GetCursorPosition()
        return x / scale, y / scale
    end

    function B.SetFrameLocation()
        local x, y = T.F.mainframe.optionbutton:GetRect()
        -- Only hide when cursor exits at the top or bottom of the option menu
        if (x < (GetScreenWidth()/2)) and (y > (GetScreenHeight()/2)) then
            if T.F.mainframe.layout ~= 1 then B.I.AnchorBagsTopLeft() end
        elseif (x < (GetScreenWidth()/2)) and (y < (GetScreenHeight()/2)) then
            if T.F.mainframe.layout ~= 2 then B.I.AnchorBagsBottomLeft() end
        elseif (x > (GetScreenWidth()/2)) and (y > (GetScreenHeight()/2)) then
            if T.F.mainframe.layout ~= 3 then B.I.AnchorBagsTopRight() end
        elseif (x > (GetScreenWidth()/2)) and (y < (GetScreenHeight()/2)) then
            if T.F.mainframe.layout ~= 4 then B.I.AnchorBagsBottomRight() end
        end
        B.ScrollbarUpdate()
    end

    function B.RowsToFit()

        local ourheight = math.ceil(T.F.mainframe.search.list:GetHeight()) - T.FRAMEWHITESPACE
        return math.floor(ourheight / T.BUTTON_HEIGHT)
    end

    function B.HideAll()

        T.F.mainframe.bank:Hide()
        T.F.mainframe.reagent:Hide()
        T.F.mainframe.void:Hide()
        T.F.mainframe.guild:Hide()
        --T.F.mainframe.search:Hide()
    end

    local ilvl_pattern = _G['ITEM_LEVEL']:gsub("%%d", "(%%d+)")
    local tooltip = CreateFrame("GameTooltip", "BW_tooltip", nil, "GameTooltipTemplate")

    function B.DetectILVL(itemlink)
        local text, ilvl

        tooltip:SetOwner(UIParent, "ANCHOR_NONE")
        tooltip:SetHyperlink(itemlink)

        text = BW_tooltipTextLeft2:GetText()
        if text then ilvl = tonumber(text:match(ilvl_pattern)) end

        tooltip:ClearLines()
        return ilvl or nil
    end

    function B.GetItemIdFromLink(itemlink)
        local itemstring = string.match(itemlink, "item[%-?%d:]+")
        local _, itemid = strsplit(":", itemstring)

        return itemid
    end

    function B.IsGuildBank(button)
        local name = button:GetName()
        if name and string.match(name, "GuildBankColumn?%dButton?%d") then
            return true
        end
        return false
    end
